from typing import ClassVar, Optional

from google.protobuf.json_format import MessageToDict
from grpc.aio import insecure_channel
from pydantic import BaseModel, validate_call

from .api_constants import (
    CONTROLLER_STUB,
    SUCCESS_CODE,
    RequestMessages,
    RPCMethodNames,
)


class ValidationError(Exception):
    pass


class ResponseCodeError(Exception):
    pass


class Discount(BaseModel):
    """
    Configure the host and port of the search server.

    Parameters:
        host (str): The host address of the search server.
        port (int | str): The port number of the search server.
    """

    host: str = "127.0.0.1"
    port: str | int = "50051"

    __instance: ClassVar = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    async def perform_operation(self, rpc_method_name, request_message):
        """Perform the operation.

        Parameters:
            rpc_method_name (str): The name of the RPC method (e.g., `Create`).
            request_message (gRPC message): The request message instance.
        """
        async with insecure_channel(f"{self.host}:{self.port}") as channel:
            CONTROLLER_STUB.channel = channel
            CONTROLLER_STUB.ready()
            RPCMethod = getattr(CONTROLLER_STUB, rpc_method_name, None)

            if RPCMethod:
                response = await RPCMethod(request_message)

                # Clear the request message for future use
                request_message.Clear()

                # Raise error if response is not successful
                if response.code != SUCCESS_CODE:
                    raise ResponseCodeError(response.code)

                return MessageToDict(response, preserving_proto_field_name=True)
            else:
                raise ValidationError("Invalid RPC method name.")

    @validate_call
    async def retrieve(self, token_identifier: str):
        """Retrieve discount using its `token_identifier`."""

        # Assign attributes
        RequestMessages.RETRIEVE.token_identifier = token_identifier

        return await self.perform_operation(
            RPCMethodNames.RETRIEVE, RequestMessages.RETRIEVE
        )

    @validate_call
    async def validate(
        self,
        token_identifier: str,
        user_identifier: Optional[str] = None,
    ):
        """Validate if a discount token is valid."""

        # Assign attributes
        RequestMessages.VALIDATE.token_identifier = token_identifier
        if user_identifier:
            RequestMessages.VALIDATE.user_identifier = user_identifier

        return await self.perform_operation(
            RPCMethodNames.VALIDATE, RequestMessages.VALIDATE
        )

    @validate_call
    async def calculate_by_token(
        self,
        amount: int,
        token_identifier: str,
        user_identifier: Optional[str] = None,
    ):
        """Calculate the discounted amount for provided token"""

        # Assign attributes
        RequestMessages.CALCULATE_BY_TOKEN.amount = amount
        RequestMessages.CALCULATE_BY_TOKEN.token_identifier = token_identifier
        if user_identifier:
            RequestMessages.CALCULATE_BY_TOKEN.user_identifier = user_identifier

        return await self.perform_operation(
            RPCMethodNames.CALCULATE_BY_TOKEN, RequestMessages.CALCULATE_BY_TOKEN
        )

    @validate_call
    async def calculate_by_item(
        self,
        amount: int,
        discount_type: str,
        item_identifier: str,
    ):
        """Calculate the lowest discounted amount for an item."""

        # Assign attributes
        RequestMessages.CALCULATE_BY_ITEM.amount = amount
        RequestMessages.CALCULATE_BY_ITEM.discount_type = discount_type
        RequestMessages.CALCULATE_BY_ITEM.item_identifier = item_identifier

        return await self.perform_operation(
            RPCMethodNames.CALCULATE_BY_ITEM, RequestMessages.CALCULATE_BY_ITEM
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: discount/grpc/discount.proto
# Protobuf Python Version: 4.25.1
# pylint: skip-file
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1c\x64iscount/grpc/discount.proto\x12\x0f\x63onfig.discount\x1a\x1cgoogle/protobuf/struct.proto"`\n\x1e\x44iscountCalculateByItemRequest\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x05\x12\x15\n\rdiscount_type\x18\x02 \x01(\t\x12\x17\n\x0fitem_identifier\x18\x03 \x01(\t"g\n\x1f\x44iscountCalculateByItemResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.Struct"}\n\x1f\x44iscountCalculateByTokenRequest\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x05\x12\x18\n\x10token_identifier\x18\x02 \x01(\t\x12\x1c\n\x0fuser_identifier\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_user_identifier"h\n DiscountCalculateByTokenResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.Struct"X\n\x10\x44iscountResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.Struct"3\n\x17\x44iscountRetrieveRequest\x12\x18\n\x10token_identifier\x18\x01 \x01(\t"e\n\x17\x44iscountValidateRequest\x12\x18\n\x10token_identifier\x18\x01 \x01(\t\x12\x1c\n\x0fuser_identifier\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_user_identifier"`\n\x18\x44iscountValidateResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.Struct2\xc5\x03\n\x12\x44iscountController\x12v\n\x0f\x43\x61lculateByItem\x12/.config.discount.DiscountCalculateByItemRequest\x1a\x30.config.discount.DiscountCalculateByItemResponse"\x00\x12y\n\x10\x43\x61lculateByToken\x12\x30.config.discount.DiscountCalculateByTokenRequest\x1a\x31.config.discount.DiscountCalculateByTokenResponse"\x00\x12Y\n\x08Retrieve\x12(.config.discount.DiscountRetrieveRequest\x1a!.config.discount.DiscountResponse"\x00\x12\x61\n\x08Validate\x12(.config.discount.DiscountValidateRequest\x1a).config.discount.DiscountValidateResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "discount.grpc.discount_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_DISCOUNTCALCULATEBYITEMREQUEST"]._serialized_start = 79
    _globals["_DISCOUNTCALCULATEBYITEMREQUEST"]._serialized_end = 175
    _globals["_DISCOUNTCALCULATEBYITEMRESPONSE"]._serialized_start = 177
    _globals["_DISCOUNTCALCULATEBYITEMRESPONSE"]._serialized_end = 280
    _globals["_DISCOUNTCALCULATEBYTOKENREQUEST"]._serialized_start = 282
    _globals["_DISCOUNTCALCULATEBYTOKENREQUEST"]._serialized_end = 407
    _globals["_DISCOUNTCALCULATEBYTOKENRESPONSE"]._serialized_start = 409
    _globals["_DISCOUNTCALCULATEBYTOKENRESPONSE"]._serialized_end = 513
    _globals["_DISCOUNTRESPONSE"]._serialized_start = 515
    _globals["_DISCOUNTRESPONSE"]._serialized_end = 603
    _globals["_DISCOUNTRETRIEVEREQUEST"]._serialized_start = 605
    _globals["_DISCOUNTRETRIEVEREQUEST"]._serialized_end = 656
    _globals["_DISCOUNTVALIDATEREQUEST"]._serialized_start = 658
    _globals["_DISCOUNTVALIDATEREQUEST"]._serialized_end = 759
    _globals["_DISCOUNTVALIDATERESPONSE"]._serialized_start = 761
    _globals["_DISCOUNTVALIDATERESPONSE"]._serialized_end = 857
    _globals["_DISCOUNTCONTROLLER"]._serialized_start = 860
    _globals["_DISCOUNTCONTROLLER"]._serialized_end = 1313
# @@protoc_insertion_point(module_scope)



from .grpc_lib.discount_pb2 import (
    DiscountCalculateByItemRequest,
    DiscountCalculateByTokenRequest,
    DiscountRetrieveRequest,
    DiscountValidateRequest,
)
from .grpc_lib.discount_pb2_grpc import DiscountControllerStub

SUCCESS_CODE = "0"

CONTROLLER_STUB = DiscountControllerStub()


class RequestMessages:
    CALCULATE_BY_TOKEN = DiscountCalculateByTokenRequest()
    CALCULATE_BY_ITEM = DiscountCalculateByItemRequest()
    RETRIEVE = DiscountRetrieveRequest()
    VALIDATE = DiscountValidateRequest()


class RPCMethodNames:
    CALCULATE_BY_TOKEN = "CalculateByToken"
    CALCULATE_BY_ITEM = "CalculateByItem"
    RETRIEVE = "Retrieve"
    VALIDATE = "Validate"
